import os

from grader_service.autograding.kube.kube_grader import KubeAutogradeExecutor
from grader_service.autograding.local_grader import LocalAutogradeExecutor, LocalProcessAutogradeExecutor

c.GraderService.service_host = "127.0.0.1"
# existing directory to use as the base directory for the grader service
service_dir = os.path.expanduser("~/Documents/Work/grader_service_dir")
c.GraderService.grader_service_dir = service_dir

c.JupyterHubGroupAuthenticator.hub_api_url = "http://127.0.0.1:8081/hub/api"

# c.LocalAutogradeExecutor.relative_input_path = "convert_in"
# c.LocalAutogradeExecutor.relative_output_path = "convert_out"

assert issubclass(KubeAutogradeExecutor, LocalAutogradeExecutor)
c.RequestHandlerConfig.autograde_executor_class = LocalAutogradeExecutor
c.LocalAutogradeExecutor.timeout_func = lambda l: 20

c.KubeAutogradeExecutor.kube_context = "minikube"
c.KubeAutogradeExecutor.default_image_name = lambda l, a: "s210.dl.hpc.tuwien.ac.at/grader/grader-notebook-minimal:arm"

print("##########: concurrent tasks = 1")
c.GraderExecutor.n_concurrent_tasks = 1



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import secrets
from grader_service.orm.assignment import Assignment
import pytest
from grader_service.server import GraderServer
import json
from grader_service.orm.submission import Submission
from unittest.mock import AsyncMock, MagicMock, Mock, patch
from grader_service.autograding.local_feedback import GenerateFeedbackExecutor
from sqlalchemy.engine.base import Engine
from sqlalchemy.orm.session import Session

# Imports are important otherwise they will not be found
from ..handlers.tornado_test_utils import *
from ..handlers.db_util import insert_submission
from sqlalchemy.orm import sessionmaker

from ...orm.submission_properties import SubmissionProperties


async def test_feedback(
        default_user,
        sql_alchemy_db,
        tmpdir
):
    default_user["groups"] = ["20wle2:instructor", "21wle1:instructor", "22wle1:instructor"]
    a_id = 1
    s_id = 1

    engine: Engine = sql_alchemy_db.engine
    insert_submission(engine, a_id, default_user["name"])

    session: Session = sessionmaker(bind=engine)()
    submission = session.query(Submission).get(s_id)
    submission_properties = session.query(SubmissionProperties).get(s_id)
    assert submission is not None
    assert submission_properties is not None
    gradebook_content = '{"notebooks":{}}'
    submission_properties.properties = gradebook_content  # we do not actually run convert so the gradebook can be empty
    session.commit()

    assert submission.score is None

    service_dir = str(tmpdir.mkdir("grader_service"))

    with patch.object(GenerateFeedbackExecutor, "_run_subprocess", return_value=None):
        executor = GenerateFeedbackExecutor(service_dir, submission)
        tmpdir.mkdir("grader_service/in")
        tmpdir.mkdir("grader_service/out")
        executor.relative_input_path = "in"
        executor.base_output_path = "out"

        await executor.start()

    submission = session.query(Submission).get(s_id)
    assert submission.feedback_available



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import pytest
from grader_service.server import GraderServer

## Imports are important otherwise they will not be found
from .tornado_test_utils import *


# @pytest.mark.asyncio
async def test_version_handler(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")

    url = service_base_url + "/"
    response = await http_server_client.fetch(url, headers={"Authorization": f"Token {default_token}"})
    assert response.code == 200
    assert response.body.decode() == "1.0"


async def test_version_handler_with_specifier(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")

    url = service_base_url + "/v1/"
    response = await http_server_client.fetch(url, headers={"Authorization": f"Token {default_token}"})
    assert response.code == 200
    assert response.body.decode() == "1.0"


# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from datetime import datetime
from re import sub
import secrets
import pytest
from grader_service.server import GraderServer
import json
from grader_service.api.models.submission import Submission
from tornado.httpclient import HTTPClientError
from datetime import timezone
from .db_util import insert_submission
from unittest.mock import AsyncMock, MagicMock, Mock, patch
from grader_service.autograding.local_grader import LocalAutogradeExecutor
from grader_service.autograding.local_feedback import GenerateFeedbackExecutor

# Imports are important otherwise they will not be found
from .db_util import insert_assignments
from .tornado_test_utils import *
from grader_service.autograding.grader_executor import GraderExecutor


async def test_auto_grading(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token,
    sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3 # default user is student
    a_id = 4

    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/grading/1/auto"

    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)
    insert_submission(engine, a_id, default_user["name"])

    with patch.object(LocalAutogradeExecutor, "start", return_value=None) as start_mock:
        response = await http_server_client.fetch(
            url, method="GET", headers={"Authorization": f"Token {default_token}"}
        )
        assert response.code == 202
        submission = Submission.from_dict(json.loads(response.body.decode()))
        assert submission.id == 1

    start_mock.assert_called()


async def test_auto_grading_wrong_assignment(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token,
    sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3 # default user is student
    a_id = 3

    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)
    insert_submission(engine, a_id, default_user["name"])

    a_id = 99
    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/grading/1/auto"

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url, method="GET", headers={"Authorization": f"Token {default_token}"}
        )
    e = exc_info.value
    assert e.code == 404


async def test_auto_grading_wrong_lecture(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token,
    sql_alchemy_db,
):
    default_user["groups"] = ["20wle2:instructor", "21wle1:instructor", "22wle1:instructor"]
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3 # default user is student
    a_id = 3

    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)
    insert_submission(engine, a_id, default_user["name"])

    l_id = 99
    # default user now instructor -> passes authorization for handler
    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/grading/1/auto"

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url, method="GET", headers={"Authorization": f"Token {default_token}"}
        )
    e = exc_info.value
    assert e.code == 403


async def test_feedback(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token,
    sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3 # default user is student
    a_id = 4

    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/grading/1/feedback"

    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)
    insert_submission(engine, a_id, default_user["name"])

    with patch.object(GenerateFeedbackExecutor, "start", return_value=None) as start_mock:
        with patch.object(GraderExecutor, "submit", return_value=None) as submit_mock:
            response = await http_server_client.fetch(
                url, method="GET", headers={"Authorization": f"Token {default_token}"}
            )
            assert response.code == 202
            submission = Submission.from_dict(json.loads(response.body.decode()))
            assert submission.id == 1

    submit_mock.assert_called()


async def test_feedback_wrong_assignment(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token,
    sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3 # default user is student
    a_id = 3

    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)
    insert_submission(engine, a_id, default_user["name"])

    a_id = 99
    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/grading/1/feedback"

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url, method="GET", headers={"Authorization": f"Token {default_token}"}
        )
    e = exc_info.value
    assert e.code == 404




import pytest
from tornado.web import HTTPError
from http import HTTPStatus

from grader_service.handlers.handler_utils import parse_ids



async def test_parse_not_numerical_ids():
    with pytest.raises(HTTPError) as e:
        parse_ids(-5, "not a number")
    assert e.value.status_code == HTTPStatus.BAD_REQUEST



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
from http import HTTPStatus

from unittest import mock
from grader_service.orm.assignment import Assignment
from grader_service.orm.group import Group
from grader_service.orm.submission import Submission
from grader_service.orm.takepart import Role, Scope
import pytest
from unittest.mock import Mock
from grader_service.handlers.git.server import GitBaseHandler
from grader_service.orm.lecture import Lecture
from tornado.web import HTTPError
from .db_util import *
import os


def get_query_side_effect(lid=1, code="ivs21s", a_type="user", scope=Scope.student, group="test_group",
                          username="test_user", a_id=1):
    def query_side_effect(input):
        m = Mock()
        if input is Lecture:
            lecture = Lecture()
            lecture.id = lid
            lecture.code = code
            m.filter.return_value.one.return_value = lecture
        elif input is Assignment:
            assignment = Assignment()
            assignment.id = a_id
            assignment.type = a_type
            m.filter.return_value.one.return_value = assignment
        elif input is Role:
            role = Role()
            role.role = scope
            m.get.return_value = role
        elif input is Group:
            g = Group()
            g.name = group
            m.get.return_value = g
        elif input is Submission:
            sub = Submission()
            sub.username = username
            m.get.return_value = sub
        else:
            m.filter.return_value.one.return_value = None
        return m

    return query_side_effect


def test_git_lookup_instructor(tmpdir):
    path = "services/grader/git/iv21s/1/source"
    git_dir = str(tmpdir.mkdir("git"))

    handler_mock = Mock()
    handler_mock.request.path = path
    handler_mock.gitbase = git_dir
    handler_mock.user.name = "test_user"
    # handler_mock.session = session

    # orm mocks
    sf = get_query_side_effect(code="iv21s", a_type="group", scope=Scope.instructor)
    handler_mock.session.query = Mock(side_effect=sf)
    constructed_git_dir = GitBaseHandler.construct_git_dir(handler_mock, repo_type="source",
                                                           lecture=sf(Lecture).filter().one(),
                                                           assignment=sf(Assignment).filter().one())
    handler_mock.construct_git_dir = Mock(return_value=constructed_git_dir)

    lookup_dir = GitBaseHandler.gitlookup(handler_mock, "send-pack")

    assert os.path.exists(lookup_dir)
    assert os.path.exists(os.path.join(lookup_dir, "HEAD"))  # is git dir
    common_path = os.path.commonpath([git_dir, lookup_dir])
    created_paths = os.path.relpath(lookup_dir, common_path)
    assert created_paths == "iv21s/1/source"


def test_git_lookup_release_pull_instructor(tmpdir):
    path = "services/grader/git/iv21s/1/release"
    git_dir = str(tmpdir.mkdir("git"))

    handler_mock = Mock()
    handler_mock.request.path = path
    handler_mock.gitbase = git_dir
    handler_mock.user.name = "test_user"
    # handler_mock.session = session

    # orm mocks
    sf = get_query_side_effect(code="iv21s", a_type="user", scope=Scope.instructor)
    handler_mock.session.query = Mock(side_effect=sf)
    constructed_git_dir = GitBaseHandler.construct_git_dir(handler_mock, repo_type="release",
                                                           lecture=sf(Lecture).filter().one(),
                                                           assignment=sf(Assignment).filter().one())
    handler_mock.construct_git_dir = Mock(return_value=constructed_git_dir)

    lookup_dir = GitBaseHandler.gitlookup(handler_mock, "upload-pack")

    assert os.path.exists(lookup_dir)
    assert os.path.exists(os.path.join(lookup_dir, "HEAD"))  # is git dir
    common_path = os.path.commonpath([git_dir, lookup_dir])
    created_paths = os.path.relpath(lookup_dir, common_path)
    assert created_paths == "iv21s/1/release"


def test_git_lookup_release_push_student_error(tmpdir):
    path = "services/grader/git/iv21s/assign_1/release"
    pathlets = path.strip("/").split("/")[3:]
    git_dir = str(tmpdir.mkdir("git"))

    handler_mock = Mock()
    handler_mock.request.path = path
    handler_mock.gitbase = git_dir
    handler_mock.user.name = "test_user"
    # handler_mock.session = session

    # orm mocks
    sf = get_query_side_effect(code="iv21s", a_type="user", scope=Scope.student)
    handler_mock.session.query = Mock(side_effect=sf)
    role = sf(Role).get()

    with pytest.raises(HTTPError) as e, pytest.warns(DeprecationWarning):  # some deprecation warning in jupyter_server
        GitBaseHandler._check_git_repo_permissions(handler_mock, "send-pack", role, pathlets)
    assert e.value.status_code == 403


def test_git_lookup_source_push_student_error(tmpdir):
    path = "services/grader/git/iv21s/assign_1/source"
    pathlets = path.strip("/").split("/")[3:]
    git_dir = str(tmpdir.mkdir("git"))

    handler_mock = Mock()
    handler_mock.request.path = path
    handler_mock.gitbase = git_dir
    handler_mock.user.name = "test_user"
    # handler_mock.session = session

    # orm mocks
    sf = get_query_side_effect(code="iv21s", a_type="user", scope=Scope.student)
    handler_mock.session.query = Mock(side_effect=sf)
    role = sf(Role).get()

    with pytest.raises(HTTPError) as e, pytest.warns(DeprecationWarning):  # some deprecation warning in jupyter_server
        GitBaseHandler._check_git_repo_permissions(handler_mock, "send-pack", role, pathlets)
    assert e.value.status_code == 403


def test_git_lookup_source_push_student_error(tmpdir):
    path = "services/grader/git/iv21s/assign_1/source"
    pathlets = path.strip("/").split("/")[3:]
    git_dir = str(tmpdir.mkdir("git"))

    handler_mock = Mock()
    handler_mock.request.path = path
    handler_mock.gitbase = git_dir
    handler_mock.user.name = "test_user"
    # handler_mock.session = session

    # orm mocks
    sf = get_query_side_effect(code="iv21s", a_type="user", scope=Scope.student)
    handler_mock.session.query = Mock(side_effect=sf)
    role = sf(Role).get()

    with pytest.raises(HTTPError) as e, pytest.warns(DeprecationWarning):  # some deprecation warning in jupyter_server
        GitBaseHandler._check_git_repo_permissions(handler_mock, "send-pack", role, pathlets)
    assert e.value.status_code == 403

def mock_git_lookup(rpc: str):
    if rpc == "bad":
        return None
    else:
        return "/path/to"


def test_get_gitdir_not_found(tmpdir):
    handler_mock = Mock()
    handler_mock.request.path = "/abc"
    handler_mock.gitlookup = mock_git_lookup
    with pytest.raises(HTTPError) as e:
        GitBaseHandler.get_gitdir(handler_mock, "bad")
    assert e.value.status_code == HTTPStatus.NOT_FOUND


def test_get_gitdir(tmpdir):
    handler_mock = Mock()
    handler_mock.request.path = "/abc"
    handler_mock.gitlookup = mock_git_lookup
    path = GitBaseHandler.get_gitdir(handler_mock, "/abc")
    assert path == "/path/to"


def test_git_lookup_push_autograde_instructor_error():
    path = "services/grader/git/iv21s/assign_1/autograde"
    pathlets = path.strip("/").split("/")[3:]
    git_dir = "/tmp"

    handler_mock = Mock()
    handler_mock.request.path = path
    handler_mock.gitbase = git_dir
    handler_mock.user.name = "test_user"

    sf = get_query_side_effect(code="iv21s", scope=Scope.instructor)
    handler_mock.session.query = Mock(side_effect=sf)
    role = sf(Role).get()

    with pytest.raises(HTTPError) as e:
        GitBaseHandler._check_git_repo_permissions(handler_mock, "send-pack", role, pathlets)
    assert e.value.status_code == 403


def test_git_lookup_push_autograde_student_error():
    path = "services/grader/git/iv21s/assign_1/autograde"
    pathlets = path.strip("/").split("/")[3:]
    git_dir = "/tmp"

    handler_mock = Mock()
    handler_mock.request.path = path
    handler_mock.gitbase = git_dir
    handler_mock.user.name = "test_user"

    sf = get_query_side_effect(code="iv21s", scope=Scope.student)
    handler_mock.session.query = Mock(side_effect=sf)
    role = sf(Role).get()
    with pytest.raises(HTTPError) as e:
        GitBaseHandler._check_git_repo_permissions(handler_mock, "send-pack", role, pathlets)
    assert e.value.status_code == 403


def test_git_lookup_push_feedback_instructor_error():
    path = "services/grader/git/iv21s/assign_1/feedback"
    pathlets = path.strip("/").split("/")[3:]
    git_dir = "/tmp"

    handler_mock = Mock()
    handler_mock.request.path = path
    handler_mock.gitbase = git_dir
    handler_mock.user.name = "test_user"

    sf = get_query_side_effect(code="iv21s", scope=Scope.instructor)
    handler_mock.session.query = Mock(side_effect=sf)
    role = sf(Role).get()
    with pytest.raises(HTTPError) as e:
        GitBaseHandler._check_git_repo_permissions(handler_mock, "send-pack", role, pathlets)
    assert e.value.status_code == 403


def test_git_lookup_push_feedback_student_error():
    path = "services/grader/git/iv21s/assign_1/feedback"
    pathlets = path.strip("/").split("/")[3:]
    git_dir = "/tmp"

    handler_mock = Mock()
    handler_mock.request.path = path
    handler_mock.gitbase = git_dir
    handler_mock.user.name = "test_user"

    sf = get_query_side_effect(code="iv21s", scope=Scope.student)
    handler_mock.session.query = Mock(side_effect=sf)
    role = sf(Role).get()
    with pytest.raises(HTTPError) as e:
        GitBaseHandler._check_git_repo_permissions(handler_mock, "send-pack", role, pathlets)
    assert e.value.status_code == 403


def test_git_lookup_pull_autograde_instructor(tmpdir):
    path = "services/grader/git/iv21s/1/autograde/1"
    git_dir = str(tmpdir.mkdir("git"))

    handler_mock = Mock()
    handler_mock.request.path = path
    handler_mock.gitbase = git_dir
    handler_mock.user.name = "test_user"

    # orm mocks
    sf = get_query_side_effect(code="iv21s", a_type="user", scope=Scope.instructor)
    handler_mock.session.query = Mock(side_effect=sf)
    role_mock = Mock()
    role_mock.role = Scope.instructor
    handler_mock.get_role = Mock(return_value=role_mock)
    constructed_git_dir = GitBaseHandler.construct_git_dir(handler_mock, repo_type="autograde",
                                                           lecture=sf(Lecture).filter().one(),
                                                           assignment=sf(Assignment).filter().one(),
                                                           submission=sf(Submission).get())
    handler_mock.construct_git_dir = Mock(return_value=constructed_git_dir)

    lookup_dir = GitBaseHandler.gitlookup(handler_mock, "upload-pack")

    assert os.path.exists(lookup_dir)
    assert os.path.exists(os.path.join(lookup_dir, "HEAD"))  # is git dir
    common_path = os.path.commonpath([git_dir, lookup_dir])
    created_paths = os.path.relpath(lookup_dir, common_path)
    assert created_paths == "iv21s/1/autograde/user/test_user"


def test_git_lookup_pull_autograde_student_error():
    path = "services/grader/git/iv21s/assign_1/autograde"
    pathlets = path.strip("/").split("/")[3:]
    git_dir = "/tmp"

    handler_mock = Mock()
    handler_mock.request.path = path
    handler_mock.gitbase = git_dir
    handler_mock.user.name = "test_user"

    sf = get_query_side_effect(code="iv21s", scope=Scope.student)
    handler_mock.session.query = Mock(side_effect=sf)
    role = sf(Role).get()

    with pytest.raises(HTTPError) as e:
        GitBaseHandler._check_git_repo_permissions(handler_mock, "upload-pack", role, pathlets)
    assert e.value.status_code == 403


def test_git_lookup_pull_feedback_instructor(tmpdir):
    path = "services/grader/git/iv21s/1/feedback/1"
    git_dir = str(tmpdir.mkdir("git"))

    handler_mock = Mock()
    handler_mock.request.path = path
    handler_mock.gitbase = git_dir
    handler_mock.user.name = "test_user"

    # orm mocks
    sf = get_query_side_effect(code="iv21s", a_type="user", scope=Scope.instructor)
    handler_mock.session.query = Mock(side_effect=sf)
    constructed_git_dir = GitBaseHandler.construct_git_dir(handler_mock, repo_type="feedback",
                                                           lecture=sf(Lecture).filter().one(),
                                                           assignment=sf(Assignment).filter().one())
    handler_mock.construct_git_dir = Mock(return_value=constructed_git_dir)

    lookup_dir = GitBaseHandler.gitlookup(handler_mock, "upload-pack")

    assert os.path.exists(lookup_dir)
    assert os.path.exists(os.path.join(lookup_dir, "HEAD"))  # is git dir
    common_path = os.path.commonpath([git_dir, lookup_dir])
    created_paths = os.path.relpath(lookup_dir, common_path)
    assert created_paths == "iv21s/1/feedback/user/test_user"


def test_git_lookup_pull_feedback_student_with_valid_id(tmpdir):
    path = "services/grader/git/iv21s/1/feedback/1"
    git_dir = str(tmpdir.mkdir("git"))

    handler_mock = Mock()
    handler_mock.request.path = path
    handler_mock.gitbase = git_dir
    handler_mock.user.name = "test_user"

    # orm mocks
    sf = get_query_side_effect(code="iv21s", a_type="user", scope=Scope.student, username="test_user")
    handler_mock.session.query = Mock(side_effect=sf)
    constructed_git_dir = GitBaseHandler.construct_git_dir(handler_mock, repo_type="feedback",
                                                           lecture=sf(Lecture).filter().one(),
                                                           assignment=sf(Assignment).filter().one())
    handler_mock.construct_git_dir = Mock(return_value=constructed_git_dir)

    lookup_dir = GitBaseHandler.gitlookup(handler_mock, "upload-pack")

    assert os.path.exists(lookup_dir)
    assert os.path.exists(os.path.join(lookup_dir, "HEAD"))  # is git dir
    common_path = os.path.commonpath([git_dir, lookup_dir])
    created_paths = os.path.relpath(lookup_dir, common_path)
    assert created_paths == "iv21s/1/feedback/user/test_user"


def test_git_lookup_pull_feedback_student_with_valid_id_extra(tmpdir):
    path = "services/grader/git/iv21s/1/feedback/1/info/refs&service=git-upload-pack"
    git_dir = str(tmpdir.mkdir("git"))

    handler_mock = Mock()
    handler_mock.request.path = path
    handler_mock.gitbase = git_dir
    handler_mock.user.name = "test_user"

    # orm mocks
    sf = get_query_side_effect(code="iv21s", a_type="user", scope=Scope.student, username="test_user")
    handler_mock.session.query = Mock(side_effect=sf)
    constructed_git_dir = GitBaseHandler.construct_git_dir(handler_mock, repo_type="feedback",
                                                           lecture=sf(Lecture).filter().one(),
                                                           assignment=sf(Assignment).filter().one())
    handler_mock.construct_git_dir = Mock(return_value=constructed_git_dir)

    lookup_dir = GitBaseHandler.gitlookup(handler_mock, "upload-pack")

    assert os.path.exists(lookup_dir)
    assert os.path.exists(os.path.join(lookup_dir, "HEAD"))  # is git dir
    common_path = os.path.commonpath([git_dir, lookup_dir])
    created_paths = os.path.relpath(lookup_dir, common_path)
    assert created_paths == "iv21s/1/feedback/user/test_user"


def test_git_lookup_pull_feedback_student_with_invalid_id_error():
    path = "services/grader/git/iv21s/assign_1/feedback/1"
    pathlets = path.strip("/").split("/")[3:]
    git_dir = "/tmp"

    handler_mock = Mock()
    handler_mock.request.path = path
    handler_mock.gitbase = git_dir
    handler_mock.user.name = "test_user"

    # test that submission with id 1 comes from "other_user"
    sf = get_query_side_effect(code="iv21s", a_type="user", scope=Scope.student, username="other_user")
    handler_mock.session.query = Mock(side_effect=sf)
    role = sf(Role).get()

    with pytest.raises(HTTPError) as e:
        GitBaseHandler._check_git_repo_permissions(handler_mock, "upload-pack", role, pathlets)
    assert e.value.status_code == 403


# /services/grader/git/20wle2/Assignment%201/feedback/2/info/refs&service=git-upload-pack
def test_git_lookup_pull_feedback_student_no_id_error():
    path = "services/grader/git/iv21s/assign_1/feedback"
    pathlets = path.strip("/").split("/")[3:]
    git_dir = "/tmp"

    handler_mock = Mock()
    handler_mock.request.path = path
    handler_mock.gitbase = git_dir
    handler_mock.user.name = "test_user"

    sf = get_query_side_effect(code="iv21s", scope=Scope.student)
    handler_mock.session.query = Mock(side_effect=sf)
    role = sf(Role).get()

    with pytest.raises(HTTPError) as e:
        GitBaseHandler._check_git_repo_permissions(handler_mock, "upload-pack", role, pathlets)
    assert e.value.status_code == 403


def test_git_lookup_pull_feedback_student_no_id_error_extra():
    path = "/services/grader/git/20wle2/Assignment%201/feedback/info/refs&service=git-upload-pack"
    pathlets = path.strip("/").split("/")[3:]
    git_dir = "/tmp"

    handler_mock = Mock()
    handler_mock.request.path = path
    handler_mock.gitbase = git_dir
    handler_mock.user.name = "test_user"

    sf = get_query_side_effect(code="iv21s", scope=Scope.student)
    handler_mock.session.query = Mock(side_effect=sf)
    role = sf(Role).get()

    with pytest.raises(HTTPError) as e:
        GitBaseHandler._check_git_repo_permissions(handler_mock, "upload-pack", role, pathlets)
    assert e.value.status_code == 403


def test_git_lookup_pull_feedback_student_bad_id_error():
    path = "services/grader/git/iv21s/assign_1/feedback/abc/"
    pathlets = path.strip("/").split("/")[3:]
    git_dir = "/tmp"

    handler_mock = Mock()
    handler_mock.request.path = path
    handler_mock.gitbase = git_dir
    handler_mock.user.name = "test_user"

    sf = get_query_side_effect(code="iv21s", scope=Scope.student)
    handler_mock.session.query = Mock(side_effect=sf)
    role = sf(Role).get()

    with pytest.raises(HTTPError) as e:
        GitBaseHandler._check_git_repo_permissions(handler_mock, "upload-pack", role, pathlets)
    assert e.value.status_code == 403



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import secrets
from datetime import datetime

from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker, Session

from grader_service.orm import Lecture, Assignment, Submission
from grader_service.orm.assignment import AutoGradingBehaviour
from grader_service.orm.base import DeleteState
from grader_service.orm.submission_properties import SubmissionProperties


def insert_users(session):
    session.execute('INSERT INTO "user" ("name") VALUES ("user1")')
    session.execute('INSERT INTO "user" ("name") VALUES ("user2")')
    session.execute('INSERT INTO "user" ("name") VALUES ("user3")')
    session.execute('INSERT INTO "user" ("name") VALUES ("user4")')


def _get_lecture(name, code):
    l = Lecture()
    l.name = name
    l.code = code
    l.state = "active"
    l.deleted = "active"
    return l


def insert_lectures(session: Engine):
    session: Session = sessionmaker(session)()
    session.add(_get_lecture("lecture1", "21wle1"))
    session.add(_get_lecture("lecture1", "21sle1"))
    session.add(_get_lecture("lecture2", "20wle2"))
    session.add(_get_lecture("lecture3", "22sle3"))
    session.add(_get_lecture("lecture4", "21sle4"))
    session.commit()
    session.flush()


def _get_assignment(name, lectid, due_date, points, status):
    a = Assignment()
    a.name = name
    a.lectid = lectid
    a.duedate = datetime.fromisoformat(due_date)
    a.points = points
    a.status = status
    a.allow_files = False
    a.automatic_grading = AutoGradingBehaviour.unassisted
    a.deleted = DeleteState.active
    a.max_submissions = None
    a.properties = None
    return a


def insert_assignments(ex, lecture_id=1):
    session: Session = sessionmaker(ex)()
    session.add(_get_assignment("assignment_1", lecture_id, "2055-06-06 23:59:00.000", 20, "released"))
    session.add(_get_assignment("assignment_2", lecture_id, "2055-07-07 23:59:00.000", 10, "created"))
    session.commit()
    session.flush()
    num_inserts = 2
    return num_inserts


def _get_submission(assignment_id, username, feedback=False, score=None):
    s = Submission()
    s.date = datetime.now()
    s.auto_status = "not_graded"
    s.manual_status = "not_graded"
    s.assignid = assignment_id
    s.username = username
    s.score = score
    s.commit_hash = secrets.token_hex(20)
    s.feedback_available = feedback
    return s


def _get_submission_properties(submission_id, properties=None):
    s = SubmissionProperties(sub_id=submission_id, properties=properties)
    return s


def insert_submission(ex, assignment_id=1, username="ubuntu", feedback=False, with_properties=True, score=None):
    # TODO Allows only one submission with properties per user because we do not have the submission id
    session: Session = sessionmaker(ex)()
    session.add(_get_submission(assignment_id, username, feedback=feedback, score=score))
    session.commit()
    if with_properties:
        id = session.query(Submission).filter(Submission.assignid==assignment_id,Submission.username==username).first().id
        session.add(_get_submission_properties(id))
        session.commit()
    session.flush()


def insert_take_part(ex, lecture_id, username="ubuntu", role="student"):
    ex.execute(f'INSERT INTO "takepart" ("username","lectid","role") VALUES ("{username}",{lecture_id},"{role}")')


def insert_grading(session):
    pass



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import os
from re import S
import pytest
from grader_service import handlers  # need import to register handlers
from grader_service.registry import HandlerPathRegistry
from grader_service.server import GraderServer
from tornado_sqlalchemy import SQLAlchemy
# import alembic
from alembic import config
from alembic.command import upgrade
from .db_util import insert_assignments, insert_lectures

__all__ = ["db_test_config", "sql_alchemy_db", "app", "service_base_url", "jupyter_hub_mock_server", "default_user", "default_token"]

from ...auth.hub import JupyterHubGroupAuthenticator


@pytest.fixture(scope="function")
def db_test_config():
    cfg = config.Config(
        os.path.abspath(os.path.dirname(__file__) + "../../../alembic_test.ini")
    )
    cfg.set_main_option("script_location", os.path.abspath(os.path.dirname(__file__) + "../../../migrate"))
    yield cfg

@pytest.fixture(scope="function")
def sql_alchemy_db(db_test_config):
    db = SQLAlchemy(url="sqlite:///:memory:")
    engine = db.engine
    with engine.begin() as connection:
        db_test_config.attributes["connection"] = connection
        # downgrade(cfg, "base")
        upgrade(db_test_config, "head")
    engine = db.engine
    insert_lectures(engine)
    insert_assignments(engine)
    yield db


@pytest.fixture(scope="function")
def app(tmpdir, sql_alchemy_db):
    service_dir = str(tmpdir.mkdir("grader_service"))
    handlers = HandlerPathRegistry.handler_list()

    JupyterHubGroupAuthenticator.hub_api_url = ""
    application = GraderServer(
        grader_service_dir=service_dir,
        base_url="/services/grader",
        auth_cls=JupyterHubGroupAuthenticator,
        handlers=handlers,
        db=sql_alchemy_db,
        cookie_secret="test",
    )
    yield application


@pytest.fixture(scope="module")
def service_base_url():
    base_url = "/services/grader"
    yield base_url


@pytest.fixture(scope="function")
def jupyter_hub_mock_server(httpserver):
    def for_user(login_user: dict, token: str):
        httpserver.expect_request(f"/user").respond_with_json(
            login_user
        )
        return httpserver

    yield for_user

@pytest.fixture(scope="function")
def default_user():
    user = {
        "kind": "user",
        "name": "ubuntu",
        "admin": False,
        "groups": ["20wle2:instructor", "21wle1:student", "22wle1:instructor"],
    }
    yield user

@pytest.fixture(scope="function")
def default_token():
    token = "token"
    yield token



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from datetime import datetime
from re import sub
import secrets
import pytest
from grader_service.server import GraderServer
import json
from tornado.httpclient import HTTPClientError
from .db_util import insert_submission, insert_take_part

# Imports are important otherwise they will not be found
from .tornado_test_utils import *


async def test_get_permission(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token,
):
    default_user["groups"] = ["20wle2:instructor", "21wle1:student", "22wle1:instructor"]

    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]
    url = service_base_url + f"/permissions"

    response = await http_server_client.fetch(
        url, method="GET", headers={"Authorization": f"Token {default_token}"}
    )
    assert response.code == 200
    permissions = json.loads(response.body.decode())
    assert isinstance(permissions, list)
    assert len(permissions) == 3
    def get_scope(v):
        if v == 0: return "student"
        if v == 1: return "tutor"
        if v == 2: return "instructor"
    
    groups = {tuple(g.split(":")) for g in default_user["groups"]}
    for p in permissions:
        t = (p["lecture_code"], get_scope(p["scope"]))
        assert t in groups
        groups.remove(t)
    assert len(groups) == 0


async def test_get_permission_new_groups(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token,
):
    default_user["groups"] = ["pytest:tutor"]

    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]
    url = service_base_url + f"/permissions/"

    response = await http_server_client.fetch(
        url, method="GET", headers={"Authorization": f"Token {default_token}"}
    )
    assert response.code == 200
    permissions = json.loads(response.body.decode())
    assert isinstance(permissions, list)
    assert len(permissions) == 1
    def get_scope(v):
        if v == 0: return "student"
        if v == 1: return "tutor"
        if v == 2: return "instructor"
    
    groups = {tuple(g.split(":")) for g in default_user["groups"]}
    for p in permissions:
        t = (p["lecture_code"], get_scope(p["scope"]))
        assert t in groups
        groups.remove(t)
    assert len(groups) == 0




# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
from http import HTTPStatus

import pytest
from grader_service.server import GraderServer
import json

from grader_service.api.models.assignment import Assignment
from grader_service.api.models.lecture import Lecture
from .db_util import insert_submission
from tornado.httpclient import HTTPClientError

# Imports are important otherwise they will not be found
from .tornado_test_utils import *
from .db_util import insert_assignments


async def test_get_lectures(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    url = service_base_url + "/lectures"
    response = await http_server_client.fetch(
        url, method="GET", headers={"Authorization": f"Token {default_token}"}
    )
    assert response.code == 200
    lectures = json.loads(response.body.decode())
    assert isinstance(lectures, list)
    assert len(lectures) > 0
    [Lecture.from_dict(l) for l in lectures]  # assert no errors


async def test_get_lectures_with_some_parameter(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    url = service_base_url + "/lectures?some_param=WS21"
    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url, method="GET", headers={"Authorization": f"Token {default_token}"}
        )
    e = exc_info.value
    assert e.code == 400


async def test_post_lectures(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token,
):
    default_user["groups"] = ["pt:instructor"]  # user has to already be in group (we only activate on post)
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    url = service_base_url + "/lectures"
    get_response = await http_server_client.fetch(
        url, method="GET", headers={"Authorization": f"Token {default_token}"}
    )
    assert get_response.code == 200
    lectures = json.loads(get_response.body.decode())
    assert isinstance(lectures, list)
    assert len(lectures) == 1
    orig_len = len(lectures)

    # same code as in group of user
    pre_lecture = Lecture(
        id=-1, name="pytest_lecture", code="pt", complete=False
    )
    post_response = await http_server_client.fetch(
        url,
        method="POST",
        headers={"Authorization": f"Token {default_token}"},
        body=json.dumps(pre_lecture.to_dict()),
    )
    assert post_response.code == 201
    post_lecture = Lecture.from_dict(json.loads(post_response.body.decode()))
    assert post_lecture.id != pre_lecture.id
    assert post_lecture.name == pre_lecture.name
    assert post_lecture.code == pre_lecture.code
    assert not post_lecture.complete

    get_response = await http_server_client.fetch(
        url, method="GET", headers={"Authorization": f"Token {default_token}"}
    )
    assert get_response.code == 200
    lectures = json.loads(get_response.body.decode())
    assert len(lectures) == orig_len

async def test_post_lectures_ghost_lecture_not_found(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token,
):
    #TODO TEST DOES NOT ENTER HANDLER
    default_user["groups"] = ["pt:instructor"]
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]
    url = service_base_url + "/lectures"

    pre_lecture = Lecture(
        id=-1, name="pytest_lecture", code="abc", complete=False
    )
    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url,
            method="POST",
            headers={"Authorization": f"Token {default_token}"},
            body=json.dumps(pre_lecture.to_dict()),
        )
    e = exc_info.value
    print(e.response.error)
    assert e.code == HTTPStatus.NOT_FOUND



async def test_post_not_found(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token,
):
    #TODO TEST DOES NOT ENTER HANDLER
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]
    url = service_base_url + "/lectures"
    pre_lecture = Lecture(
        id=-1, name="pytest_lecture", code="pt", complete=False
    )
    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url,
            method="POST",
            headers={"Authorization": f"Token {default_token}"},
            body=json.dumps(pre_lecture.to_dict()),
        )
    e = exc_info.value
    assert e.code == 404


async def test_post_unknown_parameter(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token,
):
    default_user["groups"] = ["pt:instructor"]  # user has to already be in group (we only activate on post)
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    url = service_base_url + "/lectures?some_param=asdf"
    # same code not in user groups
    pre_lecture = Lecture(
        id=-1, name="pytest_lecture", code="pt", complete=False
    )
    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url,
            method="POST",
            headers={"Authorization": f"Token {default_token}"},
            body=json.dumps(pre_lecture.to_dict()),
        )
    e = exc_info.value
    assert e.code == 400


async def test_put_lecture(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]
    url = service_base_url + "/lectures/3"

    get_response = await http_server_client.fetch(
        url,
        method="GET",
        headers={"Authorization": f"Token {default_token}"},
    )
    assert get_response.code == 200
    lecture = Lecture.from_dict(json.loads(get_response.body.decode()))
    lecture.name = "new name"
    lecture.complete = not lecture.complete
    # lecture code will not be updated
    lecture.code = "some"

    put_response = await http_server_client.fetch(
        url,
        method="PUT",
        headers={"Authorization": f"Token {default_token}"},
        body=json.dumps(lecture.to_dict()),
    )
    
    assert put_response.code == 200
    put_lecture = Lecture.from_dict(json.loads(put_response.body.decode()))
    assert put_lecture.name == lecture.name
    assert put_lecture.complete == lecture.complete
    assert put_lecture.code != lecture.code


async def test_put_lecture_unauthorized(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token,
):
    default_user["groups"] = ["20wle2:student", "21wle1:student", "22wle1:student"]
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]
    url = service_base_url + "/lectures/1"

    get_response = await http_server_client.fetch(
        url,
        method="GET",
        headers={"Authorization": f"Token {default_token}"},
    )
    assert get_response.code == 200
    lecture = Lecture.from_dict(json.loads(get_response.body.decode()))
    lecture.name = "new name"

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url,
            method="PUT",
            headers={"Authorization": f"Token {default_token}"},
            body=json.dumps(lecture.to_dict()),
        )
    
    e = exc_info.value
    assert e.code == 403


async def test_get_lecture(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]
    url = service_base_url + "/lectures/1"

    get_response = await http_server_client.fetch(
        url,
        method="GET",
        headers={"Authorization": f"Token {default_token}"},
    )
    assert get_response.code == 200
    Lecture.from_dict(json.loads(get_response.body.decode()))


async def test_get_lecture_not_found(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]
    url = service_base_url + "/lectures/999"

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url,
            method="GET",
            headers={"Authorization": f"Token {default_token}"},
        )
    e = exc_info.value
    assert e.code == 403


async def test_delete_lecture(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]
    url = service_base_url + "/lectures/3"

    delete_response = await http_server_client.fetch(
        url,
        method="DELETE",
        headers={"Authorization": f"Token {default_token}"},
    )
    assert delete_response.code == 200

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url,
            method="GET",
            headers={"Authorization": f"Token {default_token}"},
        )
    e = exc_info.value
    assert e.code == 404


async def test_delete_lecture_unauthorized(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token,
):
    default_user["groups"] = ["20wle2:student", "21wle1:student", "22wle1:student"]
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]
    url = service_base_url + "/lectures/1"

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url,
            method="DELETE",
            headers={"Authorization": f"Token {default_token}"},
        )
    e = exc_info.value
    assert e.code == 403


async def test_delete_lecture_assignment_with_submissions(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token,
    sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3
    a_id = 2
    url = service_base_url + f"/lectures/{l_id}"

    engine = sql_alchemy_db.engine
    insert_assignments(engine, lecture_id=3)
    insert_submission(engine, assignment_id=a_id, username=default_user["name"])

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url,
            method="DELETE",
            headers={"Authorization": f"Token {default_token}"},
        )
    e = exc_info.value
    assert e.code == HTTPStatus.CONFLICT


async def test_delete_lecture_assignment_released(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token,
    sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3
    url = service_base_url + f"/lectures/{l_id}"

    engine = sql_alchemy_db.engine
    insert_assignments(engine, lecture_id=3)  # assignment with id 1 is status released

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url,
            method="DELETE",
            headers={"Authorization": f"Token {default_token}"},
        )
    e = exc_info.value
    assert e.code == HTTPStatus.CONFLICT


async def test_delete_lecture_assignment_complete(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token,
    sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3
    url = service_base_url + f"/lectures/{l_id}/assignments"

    pre_assignment = Assignment(id=-1, name="pytest", type="user", due_date=None, status="complete", points=None, automatic_grading="unassisted")
    post_response = await http_server_client.fetch(
        url,
        method="POST",
        headers={"Authorization": f"Token {default_token}"},
        body=json.dumps(pre_assignment.to_dict()),
    )
    assert post_response.code == 201

    url = service_base_url + f"/lectures/{l_id}"
    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url,
            method="DELETE",
            headers={"Authorization": f"Token {default_token}"},
        )
    e = exc_info.value
    assert e.code == HTTPStatus.CONFLICT

async def test_delete_lecture_not_found(
    app: GraderServer,
    service_base_url,
    http_server_client,
    jupyter_hub_mock_server,
    default_user,
    default_token,
    sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = -5

    url = service_base_url + f"/lectures/{l_id}"
    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url,
            method="DELETE",
            headers={"Authorization": f"Token {default_token}"},
        )
    e = exc_info.value
    assert e.code == HTTPStatus.NOT_FOUND




import pytest
from grader_service.migrate.migrate import DB_URL_PATT, get_matching_config

@pytest.fixture
def valid_db_urls():
    return [
        "c.GraderService.db_url = 'postgresql://username:password@localhost:5432/mydatabase'",
        "c.GraderService.db_url = 'postgresql://user:pass@myhost:8888/mydb'",
        "c.GraderService.db_url = 'postgresql://user:pass@localhost:5432/mydatabase/subpath'",
        "c.GraderService.db_url = 'postgresql://user:pass@localhost:5432/mydb?sslmode=require&sslcert=/path/to/cert.pem'",
        "c.GraderService.db_url = 'postgresql://user:pass@localhost:5432/mydb?connect_timeout=10&application_name=MyApp'",
        "c.GraderService.db_url = 'sqlite:////path/to/database.db'",
        "c.GraderService.db_url = 'sqlite:///relative/path/database.db'",
        "c.GraderService.db_url = 'sqlite:///C:/path/to/database.db'",
        "c.GraderService.db_url = 'sqlite:///:memory:'",
        "c.GraderService.db_url = 'sqlite:///path/to/database.db?timeout=5000&mode=ro'",
        "c.GraderService.db_url = 'postgresql://user:pass@localhost:5432/mydb?dummy1=value1&dummy2=value2'",
        "c.GraderService.db_url = 'postgresql://%75%73%65%72:%70%61%73%73@localhost:5432/mydb'",
        "c.GraderService.db_url = 'postgresql://user:pass@[2001:db8:85a3::8a2e:370:7334]:5432/mydb'",
        "c.GraderService.db_url = 'sqlite:///path/to/database.db?param1=value1&param2=value2'",
        "c.GraderService.db_url = 'sqlite:///path/to/database.db?timeout=5000&mode=ro&param1=value1'",
        "c.GraderService.db_url = 'sqlite:///path/to/unicode_路径.db'",
        "c.GraderService.db_url = 'postgresql://user:pass@localhost:5432/mydb?param1=value1&param2=value2&param3=value3&param4=value4&param5=value5'"
    ]


@pytest.fixture
def invalid_db_urls():
    return [
        "c.GraderService.db_url = 'postgresql://user:password@localhost:5432/'",  # Missing database name
        "c.GraderService.db_url = 'sqlite://localhost/dbname'",  # Missing absolute path
        "c.GraderService.db_url = 'sqlite:///path/to/database'",  # Missing file extension
        "c.GraderService.db_url = 'postgresql:user:password@localhost:5432/mydatabase'",  # Missing double slashes after "postgresql:"
        "c.GraderService.db_url = 'sqlite:/path/to/database.db'",  # Missing triple slashes after "sqlite:"
        "c.GraderService.db_url = 'postgresql://user:pass@localhost:5432/mydb?invalid_param'",  # Invalid query parameter
        "c.GraderService.db_url = 'postgresql://user:pass@localhost:5432/mydb?param1=value1&'",  # Invalid query parameter syntax
        "c.GraderService.db_url = 'sqlite:///path/to/database.db?param1=value1&param2=value2&'",  # Invalid query parameter syntax
        "c.GraderService.db_url = 'postgresql://user:pass@localhost:5432/mydb?param1=value1&param2=value2&param3'",  # Invalid query parameter syntax
        "c.GraderService.db_url = 'sqlite://localhost/dbname?param1=value1&param2=value2'",  # Invalid query parameter syntax
        "c.GraderService.db_url = 'postgresql://user:pass@localhost:5432/mydb?param1=value1?param2=value2'",  # Duplicate query parameter separator
        "c.GraderService.db_url = 'sqlite://user:pass@localhost:5432/mydb'",  # Invalid scheme for SQLite
        "c.GraderService.db_url = 'postgresql://user:pass@localhost:5432/mydb?param1=value1&param2=value2&param3=value3&param4=value4&param5=value5&'",  # Invalid query parameter syntax
        "c.GraderService.db_url = 'sqlite:///path/to/database.db?param1=value1&&param2=value2'",  # Duplicate query parameter separator
        "c.GraderService.db_url = 'postgresql://user:pass@localhost:5432/mydb?param1=value1&param2=value2&param3=value3&&param4=value4'",  # Duplicate query parameter separator
        "c.GraderService.db_url = 'sqlite://user:pass@localhost:5432/mydb?param1=value1&param2=value2&param3='",  # Invalid query parameter syntax
        "c.GraderService.db_url = 'postgresql://user:pass@localhost:5432/mydb?'",  # Empty query parameter
        "c.GraderService.db_url = 'sqlite:///path/to/database.db?'",  # Empty query parameter
    ]


def test_db_url_regex_matches_expected(valid_db_urls):
    "Checks that the regex matches the expected URLs"
    for url in valid_db_urls:
        assert get_matching_config(url, DB_URL_PATT) is not None




# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from tornado import web
from traitlets import config


class GraderServer(config.LoggingConfigurable, web.Application):
    """As an unmanage jupyter hub service, the application gets these
    environment variables from the hub

    See jupyterhub docs section on launching-a-hub-managed-service
    link: https://shorturl.at/yDGHZ"""

    def __init__(self, grader_service_dir: str, base_url: str,
                 auth_cls, **kwargs):
        super().__init__(**kwargs)
        self.grader_service_dir = grader_service_dir
        self.base_url = base_url
        self.auth_cls = auth_cls



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
import csv
import subprocess
from datetime import datetime
from re import sub
import secrets
from unittest.mock import patch

import isodate
import pytest
from grader_service.server import GraderServer
import json
from grader_service.orm.assignment import Assignment as AssignmentORM
from grader_service.api.models.submission import Submission
from tornado.httpclient import HTTPClientError
from datetime import timezone
from .db_util import insert_submission, insert_take_part, _get_assignment

# Imports are important otherwise they will not be found
from .tornado_test_utils import *
from .db_util import insert_assignments
from ...api.models.assignment import Assignment
from ...handlers.base_handler import GraderBaseHandler
from ...handlers.submissions import SubmissionHandler
from ...orm import Role
from ...orm.takepart import Scope


async def submission_test_setup(sql_alchemy_db, http_server_client, default_user, default_token,
                                url: str, a_id: int):
    engine = sql_alchemy_db.engine
    insert_submission(engine, assignment_id=a_id, username=default_user["name"])
    insert_submission(engine, assignment_id=a_id, username=default_user["name"], with_properties=False)
    # should make no difference
    insert_submission(engine, assignment_id=a_id, username="user1")
    insert_submission(engine, assignment_id=a_id, username="user1", with_properties=False)

    response = await http_server_client.fetch(
        url, method="GET", headers={"Authorization": f"Token {default_token}"}
    )
    return response


@pytest.mark.parametrize("period,expected", [("P0D", 1.0), ("P1D", 0.5), ("P2D", 0.2), ("P3D", 0.1)])
def test_calculate_late_submission_scaling(period, expected):
    a = AssignmentORM()
    role = Role()
    role.role = Scope.student
    now = datetime.utcnow()
    a.duedate = now

    settings = {
        "late_submission": [{"period": "P1D", "scaling": 0.5},
                            {"period": "P2D", "scaling": 0.2},
                            {"period": "P3D", "scaling": 0.1}]
    }
    a.settings = json.dumps(settings)

    submission_ts = now - isodate.parse_duration("PT1S") + isodate.parse_duration(period)
    assert SubmissionHandler.calculate_late_submission_scaling(a, submission_ts, role) == expected


def test_calculate_late_submission_scaling_error():
    a = AssignmentORM()
    role = Role()
    role.role = Scope.student
    now = datetime.utcnow()
    a.duedate = now

    settings = {
        "late_submission": [{"period": "P1D", "scaling": 0.5}]
    }
    a.settings = json.dumps(settings)

    submission_ts = now + isodate.parse_duration("P1M")
    from tornado.web import HTTPError
    with pytest.raises(HTTPError):
        SubmissionHandler.calculate_late_submission_scaling(a, submission_ts, role)


async def test_get_submissions(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    a_id = 1
    url = service_base_url + f"/lectures/1/assignments/{a_id}/submissions/"
    response = await submission_test_setup(sql_alchemy_db,
                                           http_server_client,
                                           default_user,
                                           default_token,
                                           url,
                                           a_id)
    assert response.code == 200
    submissions = json.loads(response.body.decode())
    assert isinstance(submissions, list)
    assert len(submissions) == 2
    assert submissions[0]["username"] == "ubuntu"
    assert submissions[1]["username"] == "ubuntu"
    Submission.from_dict(submissions[0])


async def test_get_submissions_format_csv(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    a_id = 1
    url = service_base_url + f"/lectures/1/assignments/{a_id}/submissions/?format=csv"
    response = await submission_test_setup(sql_alchemy_db,
                                           http_server_client,
                                           default_user,
                                           default_token,
                                           url,
                                           a_id)
    assert response.code == 200
    decoded_content = response.body.decode('utf-8')

    body_csv = csv.reader(decoded_content.splitlines(), delimiter=',')
    submissions = list(body_csv)
    # Delete column description
    submissions.pop(0)

    assert len(submissions) == 2
    assert submissions[0][4] == "ubuntu"
    assert submissions[1][4] == "ubuntu"
    Submission.from_dict(submissions[0])


async def test_get_submissions_format_wrong(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 1  # default user is student
    a_id = 1
    engine = sql_alchemy_db.engine

    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/?format=abc"

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url, method="GET", headers={"Authorization": f"Token {default_token}"}
        )
    e = exc_info.value
    assert e.code == 400


async def test_get_submissions_filter_wrong(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 1  # default user is student
    a_id = 1
    engine = sql_alchemy_db.engine

    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/?filter=abc"

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url, method="GET", headers={"Authorization": f"Token {default_token}"}
        )
    e = exc_info.value
    assert e.code == 400


async def test_get_submissions_instructor_version(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3
    a_id = 4
    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)

    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/?instructor-version=true"

    insert_submission(engine, assignment_id=a_id, username=default_user["name"])
    insert_submission(engine, assignment_id=a_id, username=default_user["name"], with_properties=False)
    insert_submission(engine, assignment_id=a_id, username="user1")
    insert_submission(engine, assignment_id=a_id, username="user1", with_properties=False)

    response = await http_server_client.fetch(
        url, method="GET", headers={"Authorization": f"Token {default_token}"}
    )
    assert response.code == 200
    submissions = json.loads(response.body.decode())
    assert isinstance(submissions, list)
    assert len(submissions) == 4
    user_submission = submissions[0:2]

    possible_users = {default_user["name"], "user1"}
    user_name = submissions[0]["username"]
    assert user_name in possible_users
    possible_users.remove(user_name)

    assert isinstance(user_submission, list)
    assert len(user_submission) == 2
    assert all([isinstance(s, dict) for s in user_submission])
    [Submission.from_dict(s) for s in user_submission]

    user_submission = submissions[2:4]

    user_name = user_submission[0]["username"]
    assert user_name in possible_users
    possible_users.remove(user_name)
    assert len(possible_users) == 0

    assert isinstance(user_submission, list)
    assert len(user_submission) == 2
    assert all([isinstance(s, dict) for s in user_submission])
    [Submission.from_dict(s) for s in user_submission]


async def test_get_submissions_instructor_version_unauthorized(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    default_user["groups"] = ["20wle2:student", "21wle1:student", "22wle1:student"]
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 1  # default user is student
    a_id = 1
    engine = sql_alchemy_db.engine

    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/?instructor-version=true"

    insert_submission(engine, assignment_id=a_id, username=default_user["name"])
    insert_submission(engine, assignment_id=a_id, username=default_user["name"], with_properties=False)

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url, method="GET", headers={"Authorization": f"Token {default_token}"}
        )
    e = exc_info.value
    assert e.code == 403


async def test_get_submissions_latest_instructor_version(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3
    a_id = 4
    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)

    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/?instructor-version=true&filter=latest"

    insert_submission(engine, assignment_id=a_id, username=default_user["name"])
    insert_submission(engine, assignment_id=a_id, username=default_user["name"], with_properties=False)
    insert_submission(engine, assignment_id=a_id, username="user1")
    insert_submission(engine, assignment_id=a_id, username="user1", with_properties=False)

    response = await http_server_client.fetch(
        url, method="GET", headers={"Authorization": f"Token {default_token}"}
    )
    assert response.code == 200
    submissions = json.loads(response.body.decode())
    assert isinstance(submissions, list)
    assert len(submissions) == 2
    # Submissions of first user
    user_submission = submissions[0:1]

    possible_users = {default_user["name"], "user1"}
    user_name = user_submission[0]["username"]
    assert user_name in possible_users
    possible_users.remove(user_name)

    assert isinstance(user_submission, list)
    assert len(user_submission) == 1
    assert all([isinstance(s, dict) for s in user_submission])
    [Submission.from_dict(s) for s in user_submission]

    user_submission = submissions[1:2]

    user_name = user_submission[0]["username"]
    assert user_name in possible_users
    possible_users.remove(user_name)
    assert len(possible_users) == 0

    assert isinstance(user_submission, list)
    assert len(user_submission) == 1
    assert all([isinstance(s, dict) for s in user_submission])
    [Submission.from_dict(s) for s in user_submission]


async def test_get_submissions_best_instructor_version(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3
    a_id = 4
    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)

    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/?instructor-version=true&filter=best"

    insert_submission(engine, assignment_id=a_id, username=default_user["name"], feedback=True, score=3)
    insert_submission(engine, assignment_id=a_id, username=default_user["name"], feedback=False, with_properties=False)
    insert_submission(engine, assignment_id=a_id, username="user1", score=3)
    insert_submission(engine, assignment_id=a_id, username="user1", with_properties=False)

    response = await http_server_client.fetch(
        url, method="GET", headers={"Authorization": f"Token {default_token}"}
    )
    assert response.code == 200
    submissions = json.loads(response.body.decode())
    assert isinstance(submissions, list)
    assert len(submissions) == 2
    # Submissions of first user
    user_submission = submissions[0:1]

    possible_users = {default_user["name"], "user1"}
    user_name = user_submission[0]["username"]
    assert user_name in possible_users
    possible_users.remove(user_name)

    assert isinstance(user_submission, list)
    assert len(user_submission) == 1
    assert all([isinstance(s, dict) for s in user_submission])
    [Submission.from_dict(s) for s in user_submission]


async def test_get_submissions_lecture_assignment_missmatch(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3
    a_id = 1
    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/"

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url, method="GET", headers={"Authorization": f"Token {default_token}"}
        )
    e = exc_info.value
    assert e.code == 404


async def test_get_submissions_wrong_assignment_id(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 1
    a_id = 99
    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/"

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url, method="GET", headers={"Authorization": f"Token {default_token}"}
        )
    e = exc_info.value
    assert e.code == 404


async def test_get_submission(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3  # user has to be instructor
    a_id = 4
    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)

    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/1/"

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url, method="GET", headers={"Authorization": f"Token {default_token}"},
        )
    e = exc_info.value
    assert e.code == 404

    insert_submission(engine, a_id, default_user["name"])

    response = await http_server_client.fetch(
        url, method="GET", headers={"Authorization": f"Token {default_token}"},
    )
    assert response.code == 200
    submission_dict = json.loads(response.body.decode())
    Submission.from_dict(submission_dict)


async def test_get_submission_assignment_lecture_missmatch(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3  # user has to be instructor
    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)

    a_id = 1  # assignment with a_id 1 is in l_id 1
    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/1/"

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url, method="GET", headers={"Authorization": f"Token {default_token}"},
        )
    e = exc_info.value
    assert e.code == 404


async def test_get_submission_assignment_submission_missmatch(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3  # user has to be instructor
    a_id = 4
    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)
    insert_submission(engine, a_id, default_user["name"])

    a_id = 1  # this assignment has no submissions
    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/1/"

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url, method="GET", headers={"Authorization": f"Token {default_token}"},
        )
    e = exc_info.value
    assert e.code == 404


async def test_get_submission_wrong_submission(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3  # user has to be instructor
    a_id = 3
    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)
    insert_submission(engine, a_id, default_user["name"])

    a_id = 1  # this assignment has no submissions
    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/99/"

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url, method="GET", headers={"Authorization": f"Token {default_token}"},
        )
    e = exc_info.value
    assert e.code == 404


async def test_get_submission_unauthorized(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
):
    default_user["groups"] = ["20wle2:student", "21wle1:student", "22wle1:student"]
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 1  # user is student
    a_id = 1

    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/1/"

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url, method="GET", headers={"Authorization": f"Token {default_token}"},
        )
    e = exc_info.value
    assert e.code == 404


async def test_put_submission(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3  # default user is student
    a_id = 4

    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/1/"

    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)
    insert_submission(engine, a_id, default_user["name"])

    pre_submission = Submission(id=-1, submitted_at=None, commit_hash=secrets.token_hex(20),
                                auto_status="automatically_graded", manual_status="manually_graded")
    response = await http_server_client.fetch(
        url, method="PUT", headers={"Authorization": f"Token {default_token}"},
        body=json.dumps(pre_submission.to_dict()),
    )
    assert response.code == 200
    submission_dict = json.loads(response.body.decode())
    submission = Submission.from_dict(submission_dict)
    assert submission.id == 1
    assert submission.auto_status == pre_submission.auto_status
    assert submission.manual_status == pre_submission.manual_status
    assert submission.commit_hash != pre_submission.commit_hash  # commit hash cannot be changed
    assert not submission.feedback_available
    # assert submission.submitted_at.strftime("%Y-%m-%dT%H:%M:%S.%f")[0:-3] + "Z" == pre_submission.submitted_at
    assert submission.score is None


async def test_put_submission_lecture_assignment_missmatch(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3
    a_id = 1
    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/1"

    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)
    insert_submission(engine, a_id, default_user["name"])

    now = datetime.utcnow().isoformat("T", "milliseconds") + "Z"
    pre_submission = Submission(id=-1, submitted_at=now, commit_hash=secrets.token_hex(20),
                                auto_status="automatically_graded", manual_status="manually_graded")
    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url, method="PUT", headers={"Authorization": f"Token {default_token}"},
            body=json.dumps(pre_submission.to_dict()),
        )
    e = exc_info.value
    assert e.code == 404


async def test_put_submission_assignment_submission_missmatch(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3  # user has to be instructor
    a_id = 3
    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)
    insert_submission(engine, a_id, default_user["name"])

    a_id = 1  # this assignment has no submissions
    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/1/"

    now = datetime.utcnow().isoformat("T", "milliseconds") + "Z"
    pre_submission = Submission(id=-1, submitted_at=now, commit_hash=secrets.token_hex(20),
                                auto_status="automatically_graded", manual_status="manually_graded")
    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url, method="PUT", headers={"Authorization": f"Token {default_token}"},
            body=json.dumps(pre_submission.to_dict()),
        )
    e = exc_info.value
    assert e.code == 404


async def test_put_submission_wrong_submission(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3  # user has to be instructor
    a_id = 3
    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)
    insert_submission(engine, a_id, default_user["name"])

    a_id = 1  # this assignment has no submissions
    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/99/"

    now = datetime.utcnow().isoformat("T", "milliseconds") + "Z"
    pre_submission = Submission(id=-1, submitted_at=now, commit_hash=secrets.token_hex(20),
                                auto_status="automatically_graded", manual_status="manually_graded")
    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url, method="PUT", headers={"Authorization": f"Token {default_token}"},
            body=json.dumps(pre_submission.to_dict()),
        )
    e = exc_info.value
    assert e.code == 404


async def test_post_submission(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
        tmp_path
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3  # user has to be instructor
    a_id = 3
    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)

    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/"

    now = datetime.utcnow().isoformat("T", "milliseconds") + "Z"
    pre_submission = Submission(id=-1, submitted_at=now, commit_hash=secrets.token_hex(20),
                                auto_status="automatically_graded", manual_status="manually_graded")

    with patch.object(subprocess, "run", return_value=None):
        with patch.object(GraderBaseHandler, "construct_git_dir", return_value=str(tmp_path)):
            response = await http_server_client.fetch(
                url, method="POST", headers={"Authorization": f"Token {default_token}"},
                body=json.dumps(pre_submission.to_dict()),
            )

    assert response.code == 201


async def test_post_submission_git_repo_not_found(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3  # user has to be instructor
    a_id = 3
    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)
    insert_submission(engine, a_id, default_user["name"])

    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/"

    now = datetime.utcnow().isoformat("T", "milliseconds") + "Z"
    pre_submission = Submission(id=-1, submitted_at=now, commit_hash=secrets.token_hex(20),
                                auto_status="automatically_graded", manual_status="manually_graded")
    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url, method="POST", headers={"Authorization": f"Token {default_token}"},
            body=json.dumps(pre_submission.to_dict()),
        )
    e = exc_info.value
    assert e.code == 404


async def test_post_submission_commit_hash_not_found(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3  # user has to be instructor
    a_id = 3
    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)
    insert_submission(engine, a_id, default_user["name"])

    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/"

    pre_submission = {"value": "10"}
    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url, method="POST", headers={"Authorization": f"Token {default_token}"},
            body=json.dumps(pre_submission),
        )
    e = exc_info.value
    assert e.code == 400


async def test_submission_properties(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3  # default user is student
    a_id = 4

    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/1/properties"

    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)
    insert_submission(engine, a_id, default_user["name"])

    prop = {"notebooks": {}}
    put_response = await http_server_client.fetch(
        url,
        method="PUT",
        headers={"Authorization": f"Token {default_token}"},
        body=json.dumps(prop),
    )
    assert put_response.code == 200
    get_response = await http_server_client.fetch(
        url,
        method="GET",
        headers={"Authorization": f"Token {default_token}"},
    )
    assert get_response.code == 200
    assignment_props = json.loads(get_response.body.decode())
    assert assignment_props == prop


async def test_submission_properties_not_correct(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3  # default user is student
    a_id = 4

    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/1/properties"

    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)
    insert_submission(engine, a_id, default_user["name"])

    prop = "{}"
    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url,
            method="PUT",
            headers={"Authorization": f"Token {default_token}"},
            body=json.dumps(prop)
        )
    e = exc_info.value
    assert e.code == 400


async def test_submission_properties_not_found(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3  # default user is student
    a_id = 4

    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/1/properties"

    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)
    insert_submission(engine, a_id, default_user["name"])
    insert_submission(engine, a_id, default_user["name"])
    insert_submission(engine, a_id, default_user["name"])

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url,
            method="GET",
            headers={"Authorization": f"Token {default_token}"}
        )
    e = exc_info.value
    assert e.code == 404


async def test_submission_properties_lecture_assignment_missmatch(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3
    a_id = 1
    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)
    insert_submission(engine, a_id, default_user["name"])

    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/1/properties"

    prop = {"test": "property", "value": 2, "bool": True, "null": None}

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url,
            method="PUT",
            headers={"Authorization": f"Token {default_token}"},
            body=json.dumps(prop),
        )
    e = exc_info.value
    assert e.code == 404

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url,
            method="GET",
            headers={"Authorization": f"Token {default_token}"},
        )
    e = exc_info.value
    assert e.code == 404


async def test_submission_properties_assignment_submission_missmatch(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3  # user has to be instructor
    a_id = 3
    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)
    insert_submission(engine, a_id, default_user["name"])

    a_id = 1  # this assignment has no submissions
    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/1/properties"

    prop = {"test": "property", "value": 2, "bool": True, "null": None}

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url,
            method="PUT",
            headers={"Authorization": f"Token {default_token}"},
            body=json.dumps(prop),
        )
    e = exc_info.value
    assert e.code == 404

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url,
            method="GET",
            headers={"Authorization": f"Token {default_token}"},
        )
    e = exc_info.value
    assert e.code == 404


async def test_submission_properties_wrong_submission(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3  # user has to be instructor
    a_id = 3
    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)
    insert_submission(engine, a_id, default_user["name"])

    a_id = 1  # this assignment has no submissions
    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/99/properties"

    prop = {"test": "property", "value": 2, "bool": True, "null": None}

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url,
            method="PUT",
            headers={"Authorization": f"Token {default_token}"},
            body=json.dumps(prop),
        )
    e = exc_info.value
    assert e.code == 404

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url,
            method="GET",
            headers={"Authorization": f"Token {default_token}"},
        )
    e = exc_info.value
    assert e.code == 404


async def test_submission_properties_not_found(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
):
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3  # user has to be instructor
    a_id = 3
    engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)
    insert_submission(engine, a_id, default_user["name"], with_properties=False)

    a_id = 1  # this assignment has no submissions
    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/1/properties"

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url,
            method="GET",
            headers={"Authorization": f"Token {default_token}"},
        )
    e = exc_info.value
    assert e.code == 404


async def test_max_submissions_assignment(
        app: GraderServer,
        service_base_url,
        http_server_client,
        jupyter_hub_mock_server,
        default_user,
        default_token,
        sql_alchemy_db,
        tmp_path
):
    default_user["groups"] = ["20wle2:student", "21wle1:student", "22wle1:student"]
    http_server = jupyter_hub_mock_server(default_user, default_token)
    app.auth_cls.hub_api_url = http_server.url_for("")[0:-1]

    l_id = 3
    a_id = 3

    session = sql_alchemy_db.sessionmaker()
    assignment_orm = _get_assignment("pytest", l_id, "2055-06-06 23:59:00.000", 20, "released")
    assignment_orm.max_submissions = 1
    assignment_orm.automatic_grading = "unassisted"
    session.add(assignment_orm)
    session.commit()

    assert assignment_orm.id == 3

    url = service_base_url + f"/lectures/{l_id}/assignments/{a_id}/submissions/"

    now = datetime.utcnow().isoformat("T", "milliseconds") + "Z"
    pre_submission = Submission(id=-1, submitted_at=now, commit_hash=secrets.token_hex(20),
                                auto_status="automatically_graded", manual_status="manually_graded")

    with patch.object(subprocess, "run", return_value=None):
        with patch.object(GraderBaseHandler, "construct_git_dir", return_value=str(tmp_path)):
            response = await http_server_client.fetch(
                url, method="POST", headers={"Authorization": f"Token {default_token}"},
                body=json.dumps(pre_submission.to_dict()),
            )
    assert response.code == 201

    with pytest.raises(HTTPClientError) as exc_info:
        await http_server_client.fetch(
            url, method="POST", headers={"Authorization": f"Token {default_token}"},
            body=json.dumps(pre_submission.to_dict()),
        )

    e = exc_info.value
    assert e.code == 409



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
import json
import shutil
import time

import isodate
import jwt
import os.path
import subprocess
from http import HTTPStatus
import tornado

from grader_service.plugins.lti import LTISyncGrades
from grader_service.autograding.grader_executor import GraderExecutor
from grader_service.autograding.local_feedback import GenerateFeedbackExecutor
from grader_service.handlers.handler_utils import parse_ids
from grader_service.api.models.submission import Submission as SubmissionModel
from grader_service.api.models.assignment_settings import AssignmentSettings as AssignmentSettingsModel
from grader_service.orm.assignment import AutoGradingBehaviour
from grader_service.orm.assignment import Assignment
from grader_service.orm.lecture import Lecture
from grader_service.orm.submission import Submission
from grader_service.orm.submission_logs import SubmissionLogs
from grader_service.orm.submission_properties import SubmissionProperties
from grader_service.orm.takepart import Role, Scope
from grader_service.registry import VersionSpecifier, register_handler
from sqlalchemy.sql.expression import func
from tornado.web import HTTPError
from grader_convert.gradebook.models import GradeBookModel

from grader_service.handlers.base_handler import GraderBaseHandler, authorize, \
    RequestHandlerConfig
import datetime


def remove_points_from_submission(submissions):
    for s in submissions:
        if not s.feedback_available:
            s.score = None
    return submissions


@register_handler(
    path=r'\/lectures\/(?P<lecture_id>\d*)\/assignments' +
         r'\/(?P<assignment_id>\d*)\/submissions\/?',
    version_specifier=VersionSpecifier.ALL,
)
class SubmissionHandler(GraderBaseHandler):
    """Tornado Handler class for http requests to
    /lectures/{lecture_id}/assignments/{assignment_id}/submissions.
    """

    def on_finish(self):
        # we do not close the session we just commit because we might run
        # LocalAutogradeExecutor or GenerateFeedbackExecutor in POST which
        # still need it
        pass

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int, assignment_id: int):
        """Return the submissions of an assignment.

        Two query parameter: latest, instructor-version.

        latest: only get the latest submissions of users.
        instructor-version: if true, get the submissions of all users in
        lecture if false, get own submissions.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :raises HTTPError: throws err if user is not authorized or
        the assignment was not found
        """
        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)
        self.validate_parameters("filter", "instructor-version", "format")
        submission_filter = self.get_argument("filter", "none")
        if submission_filter not in ["none", "latest", "best"]:
            raise HTTPError(HTTPStatus.BAD_REQUEST, reason="Filter parameter \
            has to be either 'none', 'latest' or 'best'")
        instr_version = self.get_argument("instructor-version", None) == "true"
        response_format = self.get_argument("format", "json")
        if response_format not in ["json", "csv"]:
            raise HTTPError(HTTPStatus.BAD_REQUEST, reason="Response format \
            can either be 'json' or 'csv'")

        role: Role = self.get_role(lecture_id)
        if instr_version and role.role < Scope.tutor:
            raise HTTPError(HTTPStatus.FORBIDDEN, reason="Forbidden")
        assignment = self.get_assignment(lecture_id, assignment_id)

        if instr_version:
            if submission_filter == 'latest':

                # build the subquery
                subquery = (
                    self.session.query(Submission.username,
                                       func.max(Submission.date).label(
                                           "max_date"))
                    .filter(Submission.assignid == assignment_id)
                    .group_by(Submission.username)
                    .subquery())

                # build the main query
                submissions = (
                    self.session.query(Submission)
                    .join(subquery,
                          (Submission.username == subquery.c.username) & (
                                  Submission.date == subquery.c.max_date) & (
                                  Submission.assignid == assignment_id))
                    .order_by(Submission.id)
                    .all())

            elif submission_filter == 'best':

                # build the subquery
                subquery = (
                    self.session.query(Submission.username,
                                       func.max(Submission.score).label(
                                           "max_score"))
                    .filter(Submission.assignid == assignment_id)
                    .group_by(Submission.username)
                    .subquery())

                # build the main query
                submissions = (
                    self.session.query(Submission)
                    .join(subquery,
                          (Submission.username == subquery.c.username) & (
                                  Submission.score == subquery.c.max_score) & (
                                  Submission.assignid == assignment_id))
                    .order_by(Submission.id)
                    .all())

            else:
                submissions = assignment.submissions
        else:
            if submission_filter == 'latest':
                # build the subquery
                subquery = (self.session.query(Submission.username,
                                               func.max(Submission.date).label(
                                                   "max_date"))
                            .filter(Submission.assignid == assignment_id)
                            .group_by(Submission.username)
                            .subquery())

                # build the main query
                submissions = (
                    self.session.query(Submission)
                    .join(subquery,
                          (Submission.username == subquery.c.username) & (
                                  Submission.date == subquery.c.max_date) & (
                                  Submission.assignid == assignment_id))
                    .filter(
                        Submission.assignid == assignment_id,
                        Submission.username == role.username, )
                    .order_by(Submission.id)
                    .all())

            elif submission_filter == 'best':

                # build the subquery
                subquery = (self.session.query(Submission.username, func.max(
                    Submission.score).label("max_score"))
                            .filter(Submission.assignid == assignment_id)
                            .group_by(Submission.username)
                            .subquery())

                # build the main query
                submissions = (
                    self.session.query(Submission)
                    .join(subquery,
                          (Submission.username == subquery.c.username) & (
                                  Submission.score == subquery.c.max_score) & (
                                  Submission.assignid == assignment_id))
                    .filter(
                        Submission.assignid == assignment_id,
                        Submission.username == role.username, )
                    .order_by(Submission.id)
                    .all())
            else:
                submissions = (
                    self.session.query(
                        Submission
                    )
                    .filter(
                        Submission.assignid == assignment_id,
                        Submission.username == role.username,
                    )
                    .order_by(Submission.id)
                    .all()
                )
        if response_format == "csv":
            # csv format does not include logs
            self.set_header("Content-Type", "text/csv")
            for i, s in enumerate(submissions):
                d = s.model.to_dict()
                if i == 0:
                    self.write(
                        ",".join((k for k in d.keys() if k != "logs")) + "\n")
                self.write(",".join(
                    (str(v) for k, v in d.items() if k != "logs")) + "\n")
        else:
            if not instr_version:
                submissions = remove_points_from_submission(submissions)

            self.write_json(submissions)
        self.session.close()  # manually close here because on_finish overwrite

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def post(self, lecture_id: int, assignment_id: int):
        """Create submission based on commit hash.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :raises HTTPError: throws err if user is not authorized or
        the assignment was not found
        """
        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)
        self.validate_parameters()
        body = tornado.escape.json_decode(self.request.body)
        try:
            commit_hash = body["commit_hash"]
        except KeyError:
            raise HTTPError(400, reason="Commit hash not found in body")

        role = self.get_role(lecture_id)
        assignment = self.get_assignment(lecture_id, assignment_id)
        if assignment.status == "complete":
            raise HTTPError(HTTPStatus.BAD_REQUEST,
                            reason="Cannot submit completed assignment!")
        if role.role == Scope.student and assignment.status != "released":
            raise HTTPError(HTTPStatus.NOT_FOUND)
        submission_ts = datetime.datetime.utcnow()

        score_scaling = 1.0
        if assignment.duedate is not None:
            score_scaling = self.calculate_late_submission_scaling(assignment, submission_ts, role)

        if assignment.max_submissions:
            submissions = assignment.submissions
            usersubmissions = [s for s in submissions if
                               s.username == role.username]
            if len(usersubmissions) >= assignment.max_submissions and role.role < Scope.tutor:
                raise HTTPError(HTTPStatus.CONFLICT, reason="Maximum number \
                of submissions reached!")

        submission = Submission()
        submission.assignid = assignment.id
        submission.date = submission_ts
        submission.username = self.user.name
        submission.feedback_available = False
        submission.score_scaling = score_scaling

        git_repo_path = self.construct_git_dir(
            repo_type=assignment.type, lecture=assignment.lecture,
            assignment=assignment)
        if git_repo_path is None or not os.path.exists(git_repo_path):
            raise HTTPError(HTTPStatus.NOT_FOUND,
                            reason="Git repository not found")


        try:
            # Commit hash "0"*40 is used to differentiate between submissions created by instructors for students and normal submissions by any user. 
            # In this case submissions for the student might not exist, so we cannot reference a non-existing commit_hash.
            # When submission is set to editted, autograder uses edit repository, so we don't need the commit_hash of the submission.
            if commit_hash != "0"*40:
                subprocess.run(
                    ["git", "branch", "main", "--contains", commit_hash],
                    cwd=git_repo_path, capture_output=True)
        except subprocess.CalledProcessError:
            raise HTTPError(HTTPStatus.NOT_FOUND, reason="Commit not found")

        submission.commit_hash = commit_hash
        submission.auto_status = "not_graded"
        submission.manual_status = "not_graded"

        automatic_grading = assignment.automatic_grading

        self.session.add(submission)
        self.session.commit()
        self.set_status(HTTPStatus.CREATED)
        self.write_json(submission)

        # If the assignment has automatic grading or fully
        # automatic grading perform necessary operations
        if automatic_grading in [AutoGradingBehaviour.auto,
                                 AutoGradingBehaviour.full_auto]:
            self.set_status(HTTPStatus.ACCEPTED)
            executor = RequestHandlerConfig.instance() \
                .autograde_executor_class(
                self.application.grader_service_dir, submission,
                close_session=False, config=self.application.config
            )
            if automatic_grading == AutoGradingBehaviour.full_auto:
                feedback_executor = GenerateFeedbackExecutor(
                    self.application.grader_service_dir, submission,
                    config=self.application.config
                )

                async def feedback_and_sync():

                    GraderExecutor.instance().submit(feedback_executor.start)
                    lecture: Lecture = self.session.query(Lecture).get(lecture_id)

                    lti_plugin = LTISyncGrades.instance()
                    data = (lecture.serialize(), assignment.serialize(), [submission.serialize()])

                    if lti_plugin.check_if_lti_enabled(*data, sync_on_feedback=True):

                        try:
                            await lti_plugin.start(*data)
                        except Exception as e:
                            self.log.error("Could not sync grades: " + str(e))

                    else:
                        self.log.info("Skipping LTI plugin as it is not enabled")

                GraderExecutor.instance().submit(executor.start, on_finish=feedback_and_sync)

                # Trigger LTI Plugin
            else:
                GraderExecutor.instance().submit(
                    executor.start,
                    lambda: self.log.info(
                        f"Autograding task of submission \
                        {submission.id} exited!")
                )
        if automatic_grading == AutoGradingBehaviour.unassisted:
            self.session.close()

    @staticmethod
    def calculate_late_submission_scaling(assignment, submission_ts, role: Role) -> float:
        assignment_settings = AssignmentSettingsModel.from_dict(json.loads(assignment.settings))
        if assignment_settings.late_submission and len(assignment_settings.late_submission) > 0:
            scaling = 0.0
            if submission_ts <= assignment.duedate:
                scaling = 1.0
            else:
                for period in assignment_settings.late_submission:
                    late_submission_date = assignment.duedate + isodate.parse_duration(period.period)
                    if submission_ts < late_submission_date:
                        scaling = period.scaling
                        break
                if scaling == 0.0 and role.role < Scope.tutor:
                    raise HTTPError(HTTPStatus.CONFLICT,
                                    reason="Submission after last late submission period of assignment!")
        else:
            if submission_ts < assignment.duedate:
                scaling = 1.0
            else:
                if role.role < Scope.tutor:
                    raise HTTPError(HTTPStatus.CONFLICT, reason="Submission after due date of assignment!")
                else:
                    scaling = 0.0
        return scaling


@register_handler(
    path=r'\/lectures\/(?P<lecture_id>\d*)\/assignments\/' +
         r'(?P<assignment_id>\d*)\/submissions\/(?P<submission_id>\d*)\/?',
    version_specifier=VersionSpecifier.ALL,
)
class SubmissionObjectHandler(GraderBaseHandler):
    """Tornado Handler class for http requests to
    /lectures/{lecture_id}/assignments/{assignment_id}/submissions
    /{submission_id}.
    """

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int, assignment_id: int,
                  submission_id: int):
        """Returns a specific submission.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :param submission_id: id of the submission
        :type submission_id: int
        """
        lecture_id, assignment_id, submission_id = parse_ids(
            lecture_id, assignment_id, submission_id
        )
        submission = self.get_submission(lecture_id, assignment_id,
                                         submission_id)
        if self.get_role(lecture_id).role == Scope.student \
                and submission.username != self.user.name:
            raise HTTPError(HTTPStatus.NOT_FOUND)
        self.write_json(submission)

    @authorize([Scope.tutor, Scope.instructor])
    async def put(self, lecture_id: int, assignment_id: int,
                  submission_id: int):
        """Updates a specific submission and returns the updated entity.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :param submission_id: id of the submission
        :type submission_id: int
        """
        lecture_id, assignment_id, submission_id = parse_ids(
            lecture_id, assignment_id, submission_id
        )
        body = tornado.escape.json_decode(self.request.body)
        sub_model = SubmissionModel.from_dict(body)
        sub = self.get_submission(lecture_id, assignment_id, submission_id)
        # sub.date = sub_model.submitted_at
        # sub.assignid = assignment_id
        
        role = self.get_role(lecture_id)
        if role.role >= Scope.instructor:
            sub.username = sub_model.username
        sub.auto_status = sub_model.auto_status
        sub.manual_status = sub_model.manual_status
        sub.edited = sub_model.edited
        sub.feedback_available = sub_model.feedback_available or False
        if sub_model.score_scaling is not None and sub.score_scaling != sub_model.score_scaling:
            sub.score_scaling = sub_model.score_scaling
            sub.score = sub_model.score_scaling * sub.grading_score
        self.session.commit()
        self.write_json(sub)


@register_handler(
    path=r'\/lectures\/(?P<lecture_id>\d*)\/assignments\/' +
         r'(?P<assignment_id>\d*)\/submissions\/(?P<submission_id>\d*)\/logs\/?',
    version_specifier=VersionSpecifier.ALL,
)
class SubmissionLogsHandler(GraderBaseHandler):
    @authorize([Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int, assignment_id: int,
                  submission_id: int):
        """Returns logs of a submission.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :param submission_id: id of the submission
        :type submission_id: int
        :raises HTTPError: throws err if the submission logs are not found
        """
        lecture_id, assignment_id, submission_id = parse_ids(
            lecture_id, assignment_id, submission_id
        )
        logs = self.session.query(SubmissionLogs).get(submission_id)
        if logs is not None:
            self.write_json(logs.logs)
        else:
            raise HTTPError(HTTPStatus.NOT_FOUND,
                            reason="Properties of submission were not found")


@register_handler(
    path=r'\/lectures\/(?P<lecture_id>\d*)\/assignments\/' +
         r'(?P<assignment_id>\d*)\/submissions\/(?P<submission_id>\d*)\/' +
         r'properties\/?',
    version_specifier=VersionSpecifier.ALL,
)
class SubmissionPropertiesHandler(GraderBaseHandler):
    """Tornado Handler class for http requests to
    /lectures/{lecture_id}/assignments/{assignment_id}/submissions/
    {submission_id}/properties.
    """

    @authorize([Scope.student, Scope.tutor, Scope.instructor])
    async def get(self, lecture_id: int, assignment_id: int,
                  submission_id: int):
        """Returns the properties of a submission,

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :param submission_id: id of the submission
        :type submission_id: int
        :raises HTTPError: throws err if the submission or
        their properties are not found
        """
        lecture_id, assignment_id, submission_id = parse_ids(
            lecture_id, assignment_id, submission_id
        )
        properties = self.session.query(SubmissionProperties).get(
            submission_id)
        if properties is not None and properties.properties is not None:
            # delete source cells from properties if user is student
            if self.get_role(lecture_id).role == Scope.student:
                model = GradeBookModel.from_dict(
                    json.loads(properties.properties))
                for notebook in model.notebooks.values():
                    notebook.source_cells_dict = {}
                self.write(json.dumps(model.to_dict()))
            else:
                self.write(properties.properties)
        else:
            raise HTTPError(HTTPStatus.NOT_FOUND,
                            reason="Properties of submission were not found")

    @authorize([Scope.tutor, Scope.instructor])
    async def put(self, lecture_id: int, assignment_id: int,
                  submission_id: int):
        """Updates the properties of a submission.

        :param lecture_id: id of the lecture
        :type lecture_id: int
        :param assignment_id: id of the assignment
        :type assignment_id: int
        :param submission_id: id of the submission
        :type submission_id: int
        :raises HTTPError: throws err if the submission are not found
        """
        lecture_id, assignment_id, submission_id = parse_ids(
            lecture_id, assignment_id, submission_id
        )
        submission = self.get_submission(lecture_id, assignment_id,
                                         submission_id)
        properties_string: str = self.request.body.decode("utf-8")

        try:
            gradebook = GradeBookModel.from_dict(json.loads(properties_string))

            score = gradebook.score
            submission.grading_score = score
            submission.score = submission.score_scaling * score

        except Exception as e:
            self.log.info(e)
            raise HTTPError(HTTPStatus.BAD_REQUEST,
                            reason="Cannot parse properties file!")

        properties = SubmissionProperties(properties=properties_string,
                                          sub_id=submission.id)

        self.session.merge(properties)

        self.session.commit()
        self.write_json(submission)

    # TODO: not used, remove?
    def get_extra_credit(self, gradebook):
        extra_credit = 0
        for notebook in gradebook.notebooks.values():
            for grades in notebook.grades:
                extra_credit += grades.extra_credit if \
                    grades.extra_credit is not None else 0
        self.log.info("Extra credit is " + str(extra_credit))
        return extra_credit


@register_handler(
    path=r'\/lectures\/(?P<lecture_id>\d*)\/assignments\/' +
         r'(?P<assignment_id>\d*)\/submissions\/(?P<submission_id>\d*)\/edit\/?',
    version_specifier=VersionSpecifier.ALL,
)
class SubmissionEditHandler(GraderBaseHandler):

    @authorize([Scope.tutor, Scope.instructor])
    async def put(self, lecture_id: int, assignment_id: int,
                  submission_id: int):
        """Create or overwrites the repository which stores changes of
        submissions files
        :param lecture_id: lecture id
        :param assignment_id: assignment id
        :param submission_id: submission id
        :return:
        """
        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)
        self.validate_parameters()

        submission = self.get_submission(lecture_id, assignment_id,
                                         submission_id)
        assignment = submission.assignment
        lecture = assignment.lecture

        # Path to repository which will store edited submission files
        git_repo_path = os.path.join(
            self.gitbase,
            lecture.code,
            str(assignment.id),
            "edit",
            str(submission_id),
        )

        # Path to repository of student which contains the submitted files
        submission_repo_path = os.path.join(
            self.gitbase,
            lecture.code,
            str(assignment.id),
            assignment.type,
            submission.username
        )

        if os.path.exists(git_repo_path):
            shutil.rmtree(git_repo_path)

        # Creating bare repository
        if not os.path.exists(git_repo_path):
            os.makedirs(git_repo_path, exist_ok=True)

        await self._run_command_async('git init --bare', git_repo_path)

        # Create temporary paths to copy the submission
        # files in the edit repository
        tmp_path = os.path.join(
            self.application.grader_service_dir,
            "tmp",
            lecture.code,
            str(assignment.id),
            "edit",
            str(submission.id),
        )

        tmp_input_path = os.path.join(
            tmp_path,
            "input"
        )

        tmp_output_path = os.path.join(
            tmp_path,
            "output"
        )

        if os.path.exists(tmp_path):
            shutil.rmtree(tmp_path, ignore_errors=True)

        os.makedirs(tmp_input_path, exist_ok=True)

        # Init local repository
        command = "git init"
        await self._run_command_async(command, tmp_input_path)

        # Pull user repository
        command = f'git pull "{submission_repo_path}" main'
        await self._run_command_async(command, tmp_input_path)
        self.log.info("Successfully cloned repo")

        # Checkout to correct submission commit
        command = f"git checkout {submission.commit_hash}"
        await self._run_command_async(command, tmp_input_path)
        self.log.info(f"Now at commit {submission.commit_hash}")

        # Copy files to output directory
        shutil.copytree(tmp_input_path, tmp_output_path,
                        ignore=shutil.ignore_patterns(".git"))

        # Init local repository
        command = "git init"
        await self._run_command_async(command, tmp_output_path)

        # Add edit remote
        command = f"git remote add edit {git_repo_path}"
        await self._run_command_async(command, tmp_output_path)
        self.log.info("Successfully added edit remote")

        # Switch to main
        command = "git switch -c main"
        await self._run_command_async(command, tmp_output_path)
        self.log.info("Successfully switched to branch main")

        # Add files to staging
        command = "git add -A"
        await self._run_command_async(command, tmp_output_path)
        self.log.info("Successfully added files to staging")

        # Commit Files
        command = 'git commit -m "Initial commit" '
        await self._run_command_async(command, tmp_output_path)
        self.log.info("Successfully commited files")

        # Push copied files
        command = "git push edit main"
        await self._run_command_async(command, tmp_output_path)
        self.log.info("Successfully pushed copied files")

        submission.edited = True
        self.session.commit()
        self.write_json(submission)


@register_handler(
    path=r"\/lectures\/(?P<lecture_id>\d*)\/assignments\/(?P<assignment_id>\d*)\/submissions\/lti\/?",
    version_specifier=VersionSpecifier.ALL,
)
class LtiSyncHandler(GraderBaseHandler):
    cache_token = {"token": None, "ttl": datetime.datetime.now()}

    @authorize([Scope.instructor])
    async def get(self, lecture_id: int, assignment_id: int):
        # build the subquery
        subquery = (self.session.query(Submission.username, func.max(Submission.date).label("max_date"))
                    .filter(Submission.assignid == assignment_id, Submission.feedback_available)
                    .group_by(Submission.username)
                    .subquery())

        # build the main query
        submissions: list[Submission] = (
            self.session.query(Submission)
            .join(subquery,
                  (Submission.username == subquery.c.username) & (Submission.date == subquery.c.max_date) & (
                          Submission.assignid == assignment_id) & Submission.feedback_available)
            .all())

        assignment: Assignment = self.get_assignment(lecture_id, assignment_id)
        lecture: Lecture = self.session.query(Lecture).get(lecture_id)

        lti_plugin = LTISyncGrades.instance()
        data = (lecture.serialize(), assignment.serialize(), [s.serialize() for s in submissions])

        if lti_plugin.check_if_lti_enabled(*data, sync_on_feedback=False):

            try:
                results = await lti_plugin.start(*data)
                self.write_json(results)
            except HTTPError as e:
                self.write_error(e.status_code, reason=e.reason)
                self.log.info("Could not sync grades: " + e.reason)
            except Exception as e:
                self.log.error("Could not sync grades: " + str(e))

        else:
            self.log.info("Skipping LTI plugin as it is not enabled")
            self.write_error(HTTPStatus.CONFLICT, reason="LTI Plugin is not enabled")



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import pytest
import json
from grader_service.handlers.base_handler import GraderBaseHandler
from grader_service.orm.assignment import Assignment, AutoGradingBehaviour
from datetime import datetime
from grader_service.api.models.error_message import ErrorMessage
from grader_service.orm.lecture import Lecture
from sqlalchemy.orm import sessionmaker

from .db_util import *
from .tornado_test_utils import *
from unittest.mock import AsyncMock, MagicMock, Mock
import asyncio


def test_string_serialization():
    assert GraderBaseHandler._serialize("test") == "test"


def test_number_serialization():
    assert GraderBaseHandler._serialize(1) == 1
    assert GraderBaseHandler._serialize(2.5) == 2.5


def test_list_serialization():
    assert GraderBaseHandler._serialize([1, 2, 3]) == [1, 2, 3]

def test_list_serialization_empty():
    assert GraderBaseHandler._serialize([]) == []

def test_tuple_serialization():
    assert GraderBaseHandler._serialize((1, 2, 3)) == (1, 2, 3)


def test_dict_serialization():
    d = {"a": 1, "b": "test", 5: 6}
    s = GraderBaseHandler._serialize(d)
    assert d == s

def test_dict_serialization_empty():
    d = {}
    s = GraderBaseHandler._serialize(d)
    assert d == s

def test_datetime_serialization():
    d = datetime.now()
    s = GraderBaseHandler._serialize(d)
    assert type(s) == str
    assert str(d) == s

def test_assignment_serialization():
    d = {
        "id": 1,
        "name": "test",
        "due_date": datetime.now(),
        "status": "created",
        'type': None,
        'points': 0,
        'automatic_grading': AutoGradingBehaviour.unassisted,
        'max_submissions': 1,
        'allow_files': None,
        'settings': {'late_submission': None}
    }
    a = Assignment(
        id=d["id"],
        name=d["name"],
        lectid=1,
        duedate=d["due_date"],
        points=0,
        status=d["status"],
        automatic_grading=d["automatic_grading"],
        max_submissions=1,
        settings='{}'
    )

    d["due_date"] = (d["due_date"].isoformat("T", "milliseconds") + "Z")
    d["automatic_grading"] = d["automatic_grading"].name
    assert GraderBaseHandler._serialize(a) == d

def test_nested_serialization():
    o = [{"b": None}, {"a": 2}, "test", {"z": []}]
    s = GraderBaseHandler._serialize(o)
    assert o == s

def test_api_model_serialization():
    err = ErrorMessage("")



# Copyright (c) 2022, TU Wien
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from grader_service.orm.assignment import Assignment
import pytest
from grader_service.orm.submission import Submission
from unittest.mock import patch
from grader_service.autograding.local_grader import LocalAutogradeExecutor
from sqlalchemy.engine.base import Engine
from sqlalchemy.orm.session import Session

# Imports are important otherwise they will not be found
from ..handlers.tornado_test_utils import default_user, sql_alchemy_db, db_test_config
from ..handlers.db_util import insert_submission, insert_assignments
from sqlalchemy.orm import sessionmaker


async def test_auto_grading(
        default_user,
        sql_alchemy_db,
        tmpdir
):
    l_id = 3  # default user is instructor
    a_id = 3
    s_id = 1

    engine: Engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)
    insert_submission(engine, a_id, default_user["name"])

    session: Session = sessionmaker(bind=engine)()
    submission = session.query(Submission).get(s_id)
    assert submission is not None
    assignment: Assignment = submission.assignment
    gradebook_content = '{"notebooks":{}}'
    assignment.properties = gradebook_content  # we do not actually run convert so the gradebook can be empty
    session.commit()

    assert submission.score is None

    service_dir = str(tmpdir.mkdir("grader_service"))

    with patch.object(LocalAutogradeExecutor, "_run_subprocess", return_value=None):
        executor = LocalAutogradeExecutor(service_dir, submission)
        tmpdir.mkdir("grader_service/in")
        tmpdir.mkdir("grader_service/out")
        executor.relative_input_path = "in"
        executor.relative_output_path = "out"

        await executor.start()

    submission = session.query(Submission).get(s_id)
    assert submission.properties.properties == gradebook_content  # because no actual autograding this is the same as assignment properties
    assert submission.auto_status == "automatically_graded"
    assert submission.score == 0  # we do not get any scores because of dummy gradebook, but it is set to 0

@pytest.mark.asyncio
async def test_auto_grading_pending(
        default_user,
        sql_alchemy_db,
        tmpdir
):
    l_id = 3  # default user is instructor
    a_id = 3
    s_id = 1

    engine: Engine = sql_alchemy_db.engine
    insert_assignments(engine, l_id)
    insert_submission(engine, a_id, default_user["name"])

    session: Session = sessionmaker(bind=engine)()
    submission = session.query(Submission).get(s_id)
    assert submission is not None
    assignment: Assignment = submission.assignment
    gradebook_content = '{"notebooks":{}}'
    assignment.properties = gradebook_content  # we do not actually run convert so the gradebook can be empty
    session.commit()

    service_dir = str(tmpdir.mkdir("grader_service"))

    with patch.object(LocalAutogradeExecutor, "_run_subprocess", return_value=None):
        executor = LocalAutogradeExecutor(service_dir, submission)
        tmpdir.mkdir("grader_service/in")
        tmpdir.mkdir("grader_service/out")
        executor.relative_input_path = "in"
        executor.relative_output_path = "out"

        await executor._pull_submission()

    submission = session.query(Submission).get(s_id)
    assert submission.auto_status == "pending"
